import mmh3
import base64
import math


class BloomFilter:
    def __init__(self, dict):
        self.k = int(dict.get('k'), 16) if type(
            dict.get('k')) == str else dict.get('k')
        self.m = int(dict.get('m'), 16) if type(
            dict.get('m')) == str else dict.get('m')
        self.base64_data = dict.get('bitset')
        self.bitset = None

    def has(self, key):
        if self.bitset is None:
            self.bitset = BitSet({'m': self.m, 'bitset': self.base64_data})
        indices = self.get_indices(key)
        for index in indices:
            if not self.bitset.has(index):
                return False
        return True

    def get_indices(self, key):
        indices = []
        base_hashes = self.get_base_hashes(key)
        for i in range(self.k):
            a = base_hashes[i % 2]
            b = (i + (i % 2)) % 4
            c = 2 + (b/2)
            d = 0xFFFFFFFFFFFFFFFF & (
                i * base_hashes[int(c)])  # truncate 64 bits
            location = 0xFFFFFFFFFFFFFFFF & (a + d)  # truncate 64 bits
            indices.append(location % self.m)
        return indices

    def get_base_hashes(self, key):
        val1, val2 = mmh3.hash64(key, 0, signed=False)
        # convert key string to bytearray
        b = bytearray()
        b.extend(key.encode())
        # append 1
        b.append(1)
        val3, val4 = mmh3.hash64(bytes(b), 0, signed=False)
        return [val1, val2, val3, val4]


class BitSet:
    def __init__(self, dict):
        # first Uint64 (i.e. 8 bytes) encodes m, next Uint64 encodes k, next Uint64 encodes m again (so we slice them out)
        decoded_bytes = base64.b64decode(dict.get('bitset'))[8*3:]
        # how many Uint64 do we need to store m bits
        array_length = math.ceil(dict.get('m') / 64)
        self.data = []
        for i in range(array_length):
            bytes_index = 8*i
            self.data.append(int.from_bytes(
                decoded_bytes[bytes_index:bytes_index+8], "big"))  # read as big endian

    def has(self, index):
        word_index = index >> 6
        words_index_i = index & (64-1)
        mask = int(math.pow(2, words_index_i))
        word = self.data[word_index]
        return (word & mask) != 0



from . import BloomFilter


def test_alert_0xb680bc4e826ee6419f738a7cd76838126a1cfd85ef4607cb0ae2b75b0a5c70af():
    k = '0xa'  # 10
    m = '0xd33'  # 3379
    content = "AAAAAAAADTMAAAAAAAAACgAAAAAAAA0zoCntpyGtcm49UqMM9AWNWqEnAxF/UO3IbAPDWRKTJq+wpI051Z+uHn7exEpOzxdllSK+ICU9OfhN+FXsLpC2HkzRYDrOJ0Zp9DG73mRPvwRTcyFHwSOwlCtNoE9pgSk8jXe2S7uTanRQrS6WnP71vagcoXHydeiJfyVIxHIEt5IydneVsPBO0ko/iypb5EtAx57IiOWgmIWqngc+mao8bY9rxe42mv4ZnG/JGRqZyxdgIWSLrpCwkqOEVyYxuHFzoH67YujJUuPsl8GY0sFMJdpuySJlEgzIELFNcsJSA9YLLgAJ7flterIjUIgTad3rgR2kNu1OQksso0NlQY2OhguEoNDrtJcWTRwYudO69Q5Kp1mcg3ax+n0iP4Q8tgjrJhD0EnLujlYQPH1Gdr8v1zJVZ2PhLVFDnXjQGoZ6Lmqi/fjOTpvHwSks+jeHuCzWF1X9Lvm7/LPWDd2XX/jmyp9df2NiA/Jb3FhZf3akolcMIo5hnXnnyn1LupGbILiIz24NwzMl3w43xOe1bDEvRKiD+s3qNhrxbmWM9V03NMYABHfM2p7W1Q=="

    b = BloomFilter({'k': k, 'm': m, 'bitset': content})

    assert b.has("0x030cb0fd022b0a66d2d6e39d0691cce86d4188b8") == True
    assert b.has("0x1727b5a84fee033c3065473cc91c23e8607eef6b") == True
    assert b.has("0x1f11417c24fecc07154ce96db1b7b6af11c3af3f") == True
    assert b.has("a") == False
    assert b.has("b") == False
    assert b.has("c") == False


def test_alert_0xd7387cfe3ad5e051ba195ab0386ae33b2359ca403d650df5d05fc6dc596db149():
    k = 11
    m = 15
    content = "AAAAAAAAAA8AAAAAAAAACwAAAAAAAAAPAAAAAAAANtI="

    b = BloomFilter({'k': k, 'm': m, 'bitset': content})

    assert b.has("0x68f180fcce6836688e9084f035309e29bf0a2095") == True
    assert b.has("a") == False
    assert b.has("b") == False
    assert b.has("c") == False
